from parser import Parser
from objects import Character, Terminal, Nonterminal, State, DFA, REG, NFA
from objects import Composition as comp
from typing import Set, Dict, Tuple


def make_dfa(states: Set[str], characters: Set[str],
             transition: Dict[Tuple[str, str], str], init: str, final: Set[str]):

    new_init = State(init)
    new_states: Set[State] = set()
    dfa = DFA(new_states.add(new_init), set(), {}, new_init, set())

    for name in states:
        state = State(name)
        dfa.states.add(state)
    for name in characters:
        character = Character(name)
        dfa.characters.add(character)
    for name in final:
        state = State(name)
        dfa.final.add(state)
    for x, y in transition:
        state = State(x)
        character = Character(y)
        move = State(transition[x, y])
        dfa.transition[state, character] = move

    dfa.check()
    return dfa


def reg_1():
    S = Nonterminal("S")
    A = Nonterminal("A")
    a = Terminal("a")
    nonterminals = {S, A}
    init = S
    terminals = {a}
    rules = dict()
    rules[S] = {(a, S), (a)}
    rules[A] = {(a)}
    reg = REG(nonterminals, terminals, rules, init)
    return reg


def reg_2():
    S = Nonterminal("S")
    A = Nonterminal("A")
    a = Terminal("a")
    nonterminals = {S, A}
    init = S
    terminals = {a}
    rules = dict()
    reg = REG(nonterminals, terminals, rules, init)
    return reg

def nfa_1():
    q0 = State("q0")
    q1 = State("q1")
    q2 = State("q2")
    a = Terminal("a")
    transition = {(q0, a): {q1, q2}, (q1, a): {q1}, (q2, a): {q2}}
    nfa = NFA({q0, q1, q2}, {a}, transition, q0, {q1, q2})
    return nfa

def main():
    dfa_1 = make_dfa({"q_0", "q_1"}, {"a"}, {("q_0", "a"): "q_1", ("q_1", "a"): "q_1"}, "q_0", {"q_1"})
    dfa_2 = make_dfa({"r_0", "r_1", "r_2"}, {"a", "b"}, {("r_0", "a"): "r_1", ("r_0", "b"): "r_1",
                                       ("r_1", "a"): "r_2", ("r_1", "b"): "r_2"}, "r_0", {"r_2"})
    dfa_3 = make_dfa({"1", "2", "3", "4", "5", "6", "7"}, {"a", "b"}, {("1", "a"): "2", ("2", "a"): "3", ("2", "b"): "4",
                    ("3", "a"): "6", ("3", "b"): "5", ("4", "a"): "3", ("4", "b"): "2", ("5", "a"): "6", ("5", "b"): "3",
                    ("6", "a"): "2", ("7", "a"): "6", ("7", "b"): "1", }, "1", {"3", "5", "6"})

    reg1 = reg_1()
    reg2 = reg_2()
    nfa1 = nfa_1()
    parser = Parser()

    print("parser.dfa_to_str():")
    print("String representation of DFA, simple/full.")
    print()
    print(parser.dfa_to_str(dfa_1))
    print()
    print(parser.dfa_to_str(dfa_1, True))
    print()

    print("______________")
    print("DFA.accepts():")
    print("Test of DFA accepting given word:")
    print("dfa_1.accepts('a'):", dfa_1.accepts("a"))
    print("dfa_1.accepts('ab'):", dfa_1.accepts("ab"))
    print()

    print("________________________________")
    print("DFA.make_total(), DFA.is_total()")
    print("Add new terminal for DFA, check its totality, then make total.")
    print()
    b = Terminal("b")
    dfa_1.characters.add(b)
    print("dfa_1.terminals.add(b), dfa_1.is_total():", dfa_1.is_total())
    print()
    total_dfa = dfa_1.total()
    print("dfa_1.total(), total_dfa.is_total():", total_dfa.is_total())
    print()
    print(parser.dfa_to_str(total_dfa, True))
    print()

    print("________________")
    print("DFA.complement()")
    print("Complement of DFA.")

    compl_dfa = dfa_1.complement()
    print(parser.dfa_to_str(compl_dfa, True))
    print()

    print("_________________")
    print("DFA.composition()")
    print("Composition makes union, intersection and subtraction of two DFAs.")

    print()
    print("second DFA:")
    print(parser.dfa_to_str(dfa_2, True))
    print()

    print("Composition:")
    union = dfa_1.composition(dfa_1, dfa_2, comp.Union)
    print(parser.dfa_to_str(union, True))
    print()
    print("Intersection:")
    intersection = dfa_1.composition(dfa_1, dfa_2, comp.Intersection)
    print(parser.dfa_to_str(intersection, True))
    print()

    print("___________________________")
    print("DFA.eliminate_unreachable()")
    print("Add new state (also to transition) and eliminate unreachable states.")

    q_x = State("q_x")
    a = Terminal("a")
    dfa_1.states.add(q_x)
    dfa_1.final.add(q_x)
    dfa_1.transition[q_x, a] = q_x
    print()
    print("DFA with new state:")
    print(parser.dfa_to_str(dfa_1, True))
    print()
    print("DFA without unreachable states:")
    eliminated = dfa_1.eliminate_unreachable()
    print(parser.dfa_to_str(eliminated, True))
    print()

    print("______________")
    print("DFA.minimize()")
    print("Minimization of DFA from slides:")
    print()
    print(parser.dfa_to_str(dfa_3, True))


    print()
    print("Minimal DFA:")
    minimal = dfa_3.minimize()
    print(parser.dfa_to_str(minimal, True))
    print()

    print("______________")
    print("DFA.canonize()")
    print("Canonization of previous minimized DFA.")
    print()
    canonical = minimal.canonize()
    print(parser.dfa_to_str(canonical, True))
    print()

    print("__________________________________")
    print("DFA.dfa_to_reg(), REG.reg_to_nfa()")
    print("Transformation of DFA to REG and REG to NFA")
    print("plus string representation of REG and NFA.")
    print()
    print("Original DFA:")
    dfa_4 = make_dfa({"q_0", "q_1"}, {"a"}, {("q_0", "a"): "q_1", ("q_1", "a"): "q_1"}, "q_0", {"q_1"})
    print(parser.dfa_to_str(dfa_4, True))
    print()
    dfareg = dfa_4.dfa_to_reg()
    print("DFA transformated to REG:")
    print(parser.reg_to_str(dfareg, True))
    print()
    regdfa = dfareg.reg_to_nfa()
    print("REG transformated to NFA:")
    print(parser.nfa_to_str(regdfa, True))
    print()

    #det = nfa1.determinize()
    #print(parser.dfa_to_str(det, True))

    print("__________________________________")
    print("DFA.is_equivalent()")
    print("Equivalence of DFAs through intersections with complements.")
    print()

    print(dfa_3.is_equivalent(canonical), ", should be True")
    print()


main()